// ================================================================================
// (c) 2010 Altera Corporation. All rights reserved.
// Altera products are protected under numerous U.S. and foreign patents, maskwork
// rights, copyrights and other intellectual property laws.
// 
// This reference design file, and your use thereof, is subject to and governed
// by the terms and conditions of the applicable Altera Reference Design License
// Agreement (either as signed by you, agreed by you upon download or as a
// "click-through" agreement upon installation andor found at www.altera.com).
// By using this reference design file, you indicate your acceptance of such terms
// and conditions between you and Altera Corporation.  In the event that you do
// not agree with such terms and conditions, you may not use the reference design
// file and please promptly destroy any copies you have made.
// 
// This reference design file is being provided on an "as-is" basis and as an
// accommodation and therefore all warranties, representations or guarantees of
// any kind (whether express, implied or statutory) including, without limitation,
// warranties of merchantability, non-infringement, or fitness for a particular
// purpose, are specifically disclaimed.  By making this reference design file
// available, Altera expressly does not recommend, suggest or require that this
// reference design file be used in combination with any other product not
// provided by Altera.
//
// S. Masepohl (smasepoh@altera.com)
// 16 Aug 2010 - Rev 1.0 Initial Release
// 9 Sept 2010 - Rev 1.1 Corrected bug in addressing. Masters must present byte
//               addressing to slaves. I was not doing this. User needs to manually
//               tell the core about the address width requirements per below
// ================================================================================
// Parameter Definitions:
// DATA_SIZE - Set this to the data width of the SPI transfer

// ADDRESS_SIZE_EXT - Set this to teh address width of the SPI plus whatever you
// need for byte addressing. Example, if DATA_SIZE is 16 and you plan on having
// 7 bits of address coming from SPI you would set the ADDRESS_SIZE_EXT to 8 and
// LEFT_SHIFT to 1. If you have DATA_SIZE of 32 and 7 bits of address coming from
// the SPI you would set the ADDRESS_SIZE_EXT to 9 and LEFT_SHIFT to 2.

// LEFT SHIFT - set as described above, this makes the provide SPI address into
// a byte address

// MSB_SELECT
module custom_spi_avalon_mm #(parameter DATA_SIZE=16,
			      ADDRESS_SIZE_EXT = 7,
			      SCLK_LEVEL_SELECT = 1,
			      LEFT_SHIFT = 1,
			      MSB_SELECT = 0
			      )
    (
     // SPI Slave Signals
     input sclk,
     input ss_n,
     input mosi,
     inout miso,
     // Avalon Memory Master Signals
     input [DATA_SIZE-1:0] avm_m0_readdata,
     input avm_m0_waitrequest,
     input clock,
     input reset_n,
     // While listed as "reg" below these are wires from other registers
     output reg [DATA_SIZE-1:0] avm_m0_writedata,
     output [ADDRESS_SIZE_EXT-1:0] avm_m0_address_ext,
     output reg avm_m0_read,
     output reg avm_m0_write			    
     );

    localparam ADDRESS_SIZE = ADDRESS_SIZE_EXT-LEFT_SHIFT;

    reg [ADDRESS_SIZE-1:0] avm_m0_address;
    
    assign avm_m0_address_ext = avm_m0_address << LEFT_SHIFT;
    
    //ceil of the log base 2
    function integer CLogB2;
	input [31:0] Depth;
	integer      i;
	begin
            i = Depth;        
        for(CLogB2 = 0; i > 0; CLogB2 = CLogB2 + 1)
        i = i >> 1;
	end
    endfunction
    
    localparam ADDR_CNT_WIDTH = CLogB2(ADDRESS_SIZE);
    localparam DATA_CNT_WIDTH = CLogB2(DATA_SIZE);
    
    localparam [3:0] IDLE                      = 4'd0,
                     WAIT_READ_OR_WRITE        = 4'd1,
                     WRITE_TRANSACTION_ADDRESS = 4'd2,
                     WRITE_TRANSACTION_DATA    = 4'd3,
                     AVALON_MM_WRITE_STATE     = 4'd4,
                     READ_TRANSACTION_ADDRESS  = 4'd5,
                     AVALON_MM_READ_STATE      = 4'd6,
                     SEND_FIRST_MISO_DATA      = 4'd7,
                     SEND_MISO_DATA            = 4'd8;
    
    reg [ADDR_CNT_WIDTH:0] addr_cnt, next_addr_cnt;
    reg [DATA_CNT_WIDTH:0] data_cnt, next_data_cnt;
    reg [ADDRESS_SIZE-1:0] avalon_mm_address, next_avalon_mm_address;
    reg [DATA_SIZE-1:0]    avalon_mm_data, next_avalon_mm_data;
    
    // Metastability Registers for asynchronous samples
    reg [2:0] sclk_ms_reg;

    reg [2:0] mosi_ms_reg,
	      ss_n_ms_reg;

    // Standard Register Declarations
    reg [DATA_SIZE-1:0] next_avm_m0_writedata;
    reg [ADDRESS_SIZE-1:0] next_avm_m0_address;
    reg next_avm_m0_read,
	next_avm_m0_write;

    reg next_tran_wait,
	tran_wait;

    reg tog_delay;
        
    // State Machine Registers and localparams
    reg [3:0] state = 4'b0000;
    reg [3:0] next_state;

    // Standard Wire Declarations    
    wire      sclk_pos_tog    = ~sclk_ms_reg[2] & sclk_ms_reg[1];
    wire      sclk_neg_tog    = sclk_ms_reg[2] & ~sclk_ms_reg[1];
    wire      mosi_data       = mosi_ms_reg[1];
    wire      ss_n_data       = ss_n_ms_reg[1];
    wire      sclk_tog_data   = (SCLK_LEVEL_SELECT == 1) ? sclk_pos_tog : sclk_neg_tog;
    wire      sclk_tog_output = tog_delay;
    // (SCLK_LEVEL_SELECT == 1) ? sclk_neg_tog : sclk_pos_tog;
    
    wire [ADDRESS_SIZE-1:0] mosi_input_data_address = ( MSB_SELECT == 1 ) ? {avalon_mm_address[ADDRESS_SIZE-2:0],mosi_data} :
			    {mosi_data,avalon_mm_address[ADDRESS_SIZE-1:1]};
                
    wire [DATA_SIZE-1:0]    mosi_input_data_data = ( MSB_SELECT == 1 ) ? {avalon_mm_data[DATA_SIZE-2:0],mosi_data} : 
			    {mosi_data,avalon_mm_data[DATA_SIZE-1:1]};
            
    wire [DATA_SIZE-1:0] miso_output_data = ( MSB_SELECT == 1 ) ? avalon_mm_data << 1 : avalon_mm_data >> 1;

        
    assign miso = ss_n ? 1'bz : (( MSB_SELECT == 1 ) ? avalon_mm_data[DATA_SIZE-1] : avalon_mm_data[0]);

    always @(posedge clock or negedge reset_n)
    begin
	if (!reset_n)
	begin
	    sclk_ms_reg <= #1 {3{1'b0}};
	    ss_n_ms_reg <= #1 {3{1'b0}};
	    mosi_ms_reg <= #1 {3{1'b0}};
	    tog_delay   <= #1 1'b0;
	end
	else
	begin
	    sclk_ms_reg <= #1 {sclk_ms_reg[1:0],sclk};
	    ss_n_ms_reg <= #1 {ss_n_ms_reg[1:0],ss_n};
	    mosi_ms_reg <= #1 {mosi_ms_reg[1:0],mosi};
	    tog_delay   <= #1  (SCLK_LEVEL_SELECT == 1) ? sclk_pos_tog : sclk_neg_tog;
	end // else: !if(!reset_n)
    end // always @ (posedge clock or negedge reset_n)

    always @(posedge clock or negedge reset_n)
    begin
	if (!reset_n)
	begin
	    state             <= #1 IDLE;
	    avalon_mm_data    <= #1 {DATA_SIZE{1'b0}};
	    avalon_mm_address <= #1 {ADDRESS_SIZE{1'b0}};
	    addr_cnt          <= #1 {ADDR_CNT_WIDTH+1{1'b0}};
	    data_cnt          <= #1 {DATA_CNT_WIDTH+1{1'b0}};
	    avm_m0_writedata  <= #1 {DATA_SIZE{1'b0}};
	    avm_m0_write      <= #1 1'b0;
	    avm_m0_read       <= #1 1'b0;
	    avm_m0_address    <= #1 {ADDRESS_SIZE{1'b0}};
	    tran_wait         <= #1 1'b0;
	end // if (!reset_n)
	else
	begin
	    state             <= #1 next_state;
	    avalon_mm_data    <= #1 next_avalon_mm_data;
	    avalon_mm_address <= #1 next_avalon_mm_address;
	    addr_cnt          <= #1 next_addr_cnt;
	    data_cnt          <= #1 next_data_cnt;
	    avm_m0_writedata  <= #1 next_avm_m0_writedata;
	    avm_m0_write      <= #1 next_avm_m0_write;
	    avm_m0_read       <= #1 next_avm_m0_read;
	    avm_m0_address    <= #1 next_avm_m0_address;
	    tran_wait         <= #1 next_tran_wait;
	end
    end // always @ (posedge clock or negedge reset_n)
    
    always @(/*AUTOSENSE*/addr_cnt or avalon_mm_address
	     or avalon_mm_data or avm_m0_address or avm_m0_read
	     or avm_m0_readdata or avm_m0_waitrequest or avm_m0_write
	     or avm_m0_writedata or data_cnt or mosi_data
	     or mosi_input_data_address or mosi_input_data_data
	     or miso_output_data or sclk_tog_data or sclk_tog_output
	     or ss_n_data or state)
    begin

	next_state = state;
	next_data_cnt = data_cnt;
	next_addr_cnt = addr_cnt;
	next_avalon_mm_address = avalon_mm_address;
	next_avalon_mm_data = avalon_mm_data;
	next_avm_m0_read = avm_m0_read;
	next_avm_m0_write = avm_m0_write;
	next_avm_m0_writedata = avm_m0_writedata;
	next_avm_m0_address = avm_m0_address;
		
	case (state)
	    
	    IDLE: begin
		next_avm_m0_read = 1'b0;
		next_avm_m0_write = 1'b0;
		next_avm_m0_writedata = {DATA_SIZE{1'b0}};
		next_avm_m0_address = {ADDRESS_SIZE{1'b0}};
		next_tran_wait = 4'b000;
		
		if (!ss_n_data & sclk_tog_data)
		begin
		    if (mosi_data)
		    begin
			next_state = WRITE_TRANSACTION_ADDRESS;
		    end
		    else
		    begin
			next_state = READ_TRANSACTION_ADDRESS;
		    end
		end // if (!ss_n_data & sclk_tog_data)
		else
		begin
		    next_state = IDLE;
		end // else: !if(!ss_n_data & sclk_tog_data)
	    end // case: IDLE

	    // This state is unreachable and is a remnant of assuming all SPI communication will have a slave select signal
	    // ss_n. If the bus does not have a slave select signal then you should connect ss_n to 1'b0
	    WAIT_READ_OR_WRITE:
	    begin
		if (sclk_tog_data)
		begin
		    if (mosi_data)
		    next_state = WRITE_TRANSACTION_ADDRESS;
		    else
		    next_state = READ_TRANSACTION_ADDRESS;
		end
	    end

	    WRITE_TRANSACTION_ADDRESS:
	    begin
		if (addr_cnt == ADDRESS_SIZE-1 & sclk_tog_data)
		begin
		    next_state = WRITE_TRANSACTION_DATA;
		    next_addr_cnt = {ADDR_CNT_WIDTH+1{1'b0}};
		    next_avalon_mm_address = mosi_input_data_address;
 		end
		else
		begin
		    next_state = WRITE_TRANSACTION_ADDRESS;
		    
		    if (sclk_tog_data)
		    begin
			next_avalon_mm_address = mosi_input_data_address;
			next_addr_cnt          = addr_cnt + 1;
		    end
		    else
		    begin
			next_avalon_mm_address = avalon_mm_address;
			next_addr_cnt          = addr_cnt;
		    end
		end // else: !if(addr_cnt == ADDR_CNT_LIMIT)
	    end // case: WRITE_TRANSACTION_ADDRESS

	    WRITE_TRANSACTION_DATA:
	    begin
		if (data_cnt == DATA_SIZE-1 & sclk_tog_data)
		begin
		    next_state    = AVALON_MM_WRITE_STATE;
		    next_data_cnt = {DATA_CNT_WIDTH+1{1'b0}};
		    next_avalon_mm_data = mosi_input_data_data;
		end
		else
		begin
		    next_state = WRITE_TRANSACTION_DATA;

		    if (sclk_tog_data)
		    begin
			next_avalon_mm_data = mosi_input_data_data;
			next_data_cnt       = data_cnt + 1;
		    end
		    else
		    begin
			next_avalon_mm_data = avalon_mm_data;
			next_data_cnt       = data_cnt;
		    end
		end // else: !if(data_cnt = DATA_CNT_LIMIT)
	    end // case: WRITE_TRANSACTION_DATA

	    AVALON_MM_WRITE_STATE:
	    begin
		next_avm_m0_address   = avalon_mm_address;
		next_avm_m0_writedata = avalon_mm_data;
		next_avm_m0_write     = 1'b1;
		next_tran_wait        = 1'b1;
		
		if (avm_m0_waitrequest)
		begin
		    next_state = AVALON_MM_WRITE_STATE;
		end
		else if (tran_wait)
		begin
		    next_state = IDLE;
		    next_avm_m0_write = 1'b0;
		end
		     else
		     begin
			 next_state = AVALON_MM_WRITE_STATE;
		     end
	    end // case: AVALON_MM_WRITE_STATE
	    
	    READ_TRANSACTION_ADDRESS:
	    begin
		if (addr_cnt == ADDRESS_SIZE-1 & sclk_tog_data)
		begin
		    next_state = AVALON_MM_READ_STATE;
		    next_avalon_mm_address = mosi_input_data_address;
		    next_addr_cnt = {ADDR_CNT_WIDTH+1{1'b0}};
		end
		else
		begin
		    next_state = READ_TRANSACTION_ADDRESS;
		    
		    if (sclk_tog_data)
		    begin
			next_avalon_mm_address = mosi_input_data_address;
			next_addr_cnt          = addr_cnt + 1;
		    end
		    else
		    begin
			next_avalon_mm_address = avalon_mm_address;
			next_addr_cnt          = addr_cnt;
		    end
		end // else: !if(addr_cnt == ADDR_CNT_LIMIT)
	    end // case: READ_TRANSACTION_ADDRESS

	    AVALON_MM_READ_STATE:
	    begin
		next_avm_m0_address = avalon_mm_address;
		next_avalon_mm_data = avm_m0_readdata;
		next_avm_m0_read    = 1'b1;
		next_tran_wait      = 1'b1;
		
		if (avm_m0_waitrequest)
		next_state = AVALON_MM_READ_STATE;
		else if (tran_wait)
		begin
		    next_state = SEND_FIRST_MISO_DATA;
		    next_avm_m0_read = 1'b0;
		end
		     else
		     begin
			 next_state = AVALON_MM_READ_STATE;
		     end
	    end // case: AVALON_MM_READ_STATE

	    SEND_FIRST_MISO_DATA:
	    begin
		if (sclk_tog_data)
		begin
		    next_state = SEND_MISO_DATA;
		    next_avalon_mm_data = avalon_mm_data;
		    next_data_cnt = data_cnt + 1;
		end
		else
		begin
		    next_state = SEND_FIRST_MISO_DATA;
		    next_avalon_mm_data = avalon_mm_data;
		    next_data_cnt = data_cnt;
		end
	    end // case: SEND_FIRST_MISO_DATA
	    
	    SEND_MISO_DATA:
	    begin
		if (data_cnt == DATA_SIZE)
		begin
		    if (sclk_tog_data)
		    begin
			next_state = IDLE;
			next_data_cnt = {DATA_CNT_WIDTH+1{1'b0}};
		    end
		    else
		    begin
			next_state    = SEND_MISO_DATA;
			next_data_cnt = data_cnt;
		    end
		end // if (data_cnt == DATA_SIZE)
		else
		begin
		    if (sclk_tog_output)
		    begin
			next_state = SEND_MISO_DATA;
			next_avalon_mm_data = miso_output_data;
			next_data_cnt = data_cnt + 1;
		    end
		    else
		    begin
			next_state = SEND_MISO_DATA;
			next_avalon_mm_data = avalon_mm_data;
			next_data_cnt = data_cnt;
		    end
		end // else: !if(data_cnt == DATA_SIZE-1)
	    end // case: SEND_MISO_DATA
	    	    
	endcase // case (state)
    end // always @ (sclk_tog_data or sclk_tog_output or mosi_data or ss_n_data)


    
endmodule // custom_spi_avalon_mm



		